 /***************************************************************/
/***          Automatically generated support file             ***/
 /***************************************************************/

 /* This file is automatically generated by 'files.py'.
    DO NOT MODIFY. Changes will be overwritten ! */


  /* internal part for psyco.c */
#if ALL_STATIC
# include "iprocessor.c"
# include "idispatcher.c"
# include "iencoding.c"
# include "ipyencoding.c"
# include "ivm-insns.c"
#else /* if !ALL_STATIC */
#endif /* !ALL_STATIC */

PSY_INLINE void initialize_processor_files(void) {
}
